package cn.com.qmth.examcloud.service.core.service;

import static org.springframework.data.domain.ExampleMatcher.GenericPropertyMatchers.contains;

import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.commons.lang.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Example;
import org.springframework.data.domain.ExampleMatcher;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;

import cn.com.qmth.examcloud.common.uac.AccessCtrlUtil;
import cn.com.qmth.examcloud.common.uac.entity.AccessUser;
import cn.com.qmth.examcloud.common.uac.enums.RoleMeta;
import cn.com.qmth.examcloud.common.util.RedisUtil;
import cn.com.qmth.examcloud.service.core.dto.UserInfo;
import cn.com.qmth.examcloud.service.core.entity.Org;
import cn.com.qmth.examcloud.service.core.entity.Student;
import cn.com.qmth.examcloud.service.core.entity.User;
import cn.com.qmth.examcloud.service.core.entity.UserLogin;
import cn.com.qmth.examcloud.service.core.entity.UserRole;
import cn.com.qmth.examcloud.service.core.enums.UserScope;
import cn.com.qmth.examcloud.service.core.enums.UserType;
import cn.com.qmth.examcloud.service.core.params.UserParam;
import cn.com.qmth.examcloud.service.core.repo.OrgRepo;
import cn.com.qmth.examcloud.service.core.repo.StudentRepo;
import cn.com.qmth.examcloud.service.core.repo.UserLoginRepo;
import cn.com.qmth.examcloud.service.core.repo.UserRepo;
import cn.com.qmth.examcloud.service.core.repo.UserRoleRepo;

/**
 * 用户服务类
 * Created by songyue on 17/1/13.
 */
@Service
public class UserService {

    @Autowired
    UserRepo userRepo;

    @Autowired
    StudentRepo studentRepo;

    @Autowired
    OrgRepo orgRepo;

    @Autowired
    UserRoleRepo userRoleRepo;

    @Autowired
    UserLoginRepo userLoginRepo;

    public Page<User> findAll(User userCriteria, Pageable pageable){
        ExampleMatcher exampleMatcher = ExampleMatcher.matching()
                .withMatcher("name", contains())
                .withMatcher("loginName", contains()).withIgnoreNullValues();
        Example<User> example = Example.of(userCriteria, exampleMatcher);
        Page<User> users = userRepo.findAll(example, pageable);
        return users;
    }

    /**
     * 初始化密码
     * @param userId
     */
    public void initPassword(long userId){
        User user = userRepo.findOne(userId);
        //根据用户类型采用不同的初始化策略
        if(user.getType() == UserType.NOT_STUDENT){
            //初始化为默认密码
            user.setPassword(UserParam.DEFAULT_PASSWORD);

        }else if(user.getType() == UserType.STUDENT){
            //截取身份证后6位为学生登录密码
            Student student = studentRepo.findByUserId(userId);
            if(StringUtils.isNotEmpty(student.getIdentityNumber())){
                user.setPassword(StringUtils.substring(student.getIdentityNumber(),-6,-1));
            }else{
                user.setPassword(UserParam.DEFAULT_PASSWORD);
            }
        }
        user.setUpdateTime(new Date());
        userRepo.save(user);
    }


    /**
     * 一般登录
     * @param loginName
     * @param password
     * @return
     */
    public UserInfo login(String loginName,
                          String password)throws Exception{
        User user = userRepo.findByLoginName(loginName);
        return loginProcess(user,password);

    }

    /**
     * 二级登录
     * @param orgId
     * @param loginName
     * @param password
     * @return
     */
    public UserInfo login(long orgId,
                          String loginName,
                          String password)throws Exception{
        User user = userRepo.findByRootOrgIdAndLoginName(orgId,loginName);
        return loginProcess(user,password);
    }

    /**
     * 登录处理
     * @param user
     * @param password
     * @return
     */
    public UserInfo loginProcess(User user, String password)throws Exception{
        if(user == null){
        	throw new RuntimeException("该用户不存在");
        }else if(!user.getPassword().equals(password)){
        	throw new RuntimeException("密码错误");
        }else if(!user.getEnable()){
            throw new RuntimeException("该用户被禁用");
        }else{
            initUserLogin(user);
            String token = AccessCtrlUtil.buildToken();
            createAccessUser(token,user,null);
            createUserLogin(token,user);
            return getUserInfo(user,token);
        }
    }

    /**
     * 初始化用户登录
     * @param user
     */
    public void initUserLogin(User user){
        //判断是否已登录,若已登录则强制已登录用户退出
        UserLogin userLogin= userLoginRepo.findFirstByUserId(user.getId());
        if(userLogin != null){
            RedisUtil.deleteByte(userLogin.getToken());
            userLoginRepo.deleteByUserId(userLogin.getUserId());
        }
    }

    /**
     * 创建用户登录记录
     * @param token
     * @param user
     */
    public void createUserLogin(String token,User user){
        UserLogin userLogin = new UserLogin();
        userLogin.setLoginTime(new Date());
        userLogin.setUserId(user.getId());
        userLogin.setToken(token);
        userLoginRepo.save(userLogin);
    }

    /**
     * 创建访问用户并加入K/V缓存
     * @param token
     * @param user
     */
    public void createAccessUser(String token,User user,Long studentId){
        AccessUser accessUser = new AccessUser();
        Set<UserRole> userRoleSet = new HashSet<UserRole>(user.getUserRoles());
        Map<String,Set<String>> rolesMap = userRoleSet.stream()
                .collect(Collectors.groupingBy(UserRole::getAppCode,
                        Collectors.mapping(UserRole::getRoleCode,Collectors.toSet())));
        accessUser.setLoginName(user.getLoginName());
        accessUser.setName(user.getName());
        accessUser.setOrgId(user.getOrgId());
        accessUser.setRootOrgId(user.getRootOrgId());
        accessUser.setUserId(user.getId());
        accessUser.setRoles(rolesMap);
        accessUser.setToken(token);
        accessUser.setStudentId(studentId);
        RedisUtil.setByte(token,accessUser);
    }

    /**
     * 获取用户信息DTO
     * @param user
     * @param token
     * @return
     */
    public UserInfo getUserInfo(User user,String token){
        UserInfo userInfo = new UserInfo();
        Org org = orgRepo.findOne(user.getOrgId());
        Org rootOrg = orgRepo.findOne(user.getRootOrgId());
        userInfo.setUserId(user.getId());
        userInfo.setOrgId(user.getOrgId());
        userInfo.setRootOrgId(user.getRootOrgId());
        userInfo.setName(user.getName());
        userInfo.setAvatar(user.getAvatar());
        userInfo.setLoginName(user.getLoginName());
        userInfo.setOrgName(org.getName());
        userInfo.setRootOrgLogo(rootOrg.getLogo());
        userInfo.setRootOrgName(rootOrg.getName());
        userInfo.setToken(token);
        userInfo.setType(user.getType().toString());
        userInfo.setUserRoles(user.getUserRoles());
        userInfo.setRoleNames(getRoleNames(user.getUserRoles()));
        return userInfo;
    }

    /**
     * 获取角色名称
     * @param userRoles
     * @return
     */
    public List<String> getRoleNames(List<UserRole> userRoles){

        List<String> roleNameList = userRoles.stream()
                .map(userRole -> RoleMeta.valueOf(userRole.getRoleCode()).getCnName())
                .collect(Collectors.toList());
        return roleNameList;

    }

    /**
     * 登出
     * @param accessUser
     */
	public void logout(AccessUser accessUser) {
		RedisUtil.deleteByte(accessUser.getToken());
        userLoginRepo.deleteByUserId(accessUser.getUserId());
	}

	public List<User> getMarker(Long rootOrgId) {
		return userRepo.findMarkerByRootOrgId(rootOrgId);
	}

	public User save(User user) throws Exception{
        user.setScope(UserScope.ORG);
        user.setType(UserType.NOT_STUDENT);
        user.setCreateTime(new Date());
		checkLoginName(user.getRootOrgId(), user.getLoginName());
		return userRepo.save(user);
	}

	private void checkLoginName(Long rootOrgId,String loginName) {
		User old = userRepo.findByRootOrgIdAndLoginName(rootOrgId,loginName);
		if(old!=null){
			throw new RuntimeException("用户名已存在");
		}
	}

	public User update(Long id, User user) {
		User old =  userRepo.findOne(id);
		if(!old.getLoginName().equals(user.getLoginName())){
			checkLoginName(user.getRootOrgId(), user.getLoginName());
		}
		user.setUpdateTime(new Date());
		return userRepo.save(user);
	}

}
